/*
 * stbadmin module
 *
 * author: Laser
 *
 *
 *
 *
 *
 */
#include "boa.h"
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <sys/signal.h>
#include <stdarg.h>

extern int isblank(int c);
extern char *nvram_get(const char *name);
extern int nvram_set(const char *name, const char *value);
extern int nvram_commit(void);

#define ARGV(args...) ((char *[]) { args, NULL })
#define ETHER_ADDR_LEN 6
#define nvram_safe_get(name) \
	(nvram_get(name) ? : "")

int echo(char *fmt, ...);

/*
 * check whether the value is in the strings argv[]
 * params:
 *   value  -- a string name
 *   argv[] -- a strings array, ex. {"hello", "goodbye", NULL}, the last one must be NULL.
 * return:
 *   0 -- not
 *   1 -- yes
 */
inline int
valid_choice(char *value, char *argv[])
{
	char **choice = argv;
	
	for (choice = argv; *choice ; choice++) {
		if (!strcmp(value, *choice))
			return 1;
	}
	
	return 0;
}

/*
 * check whether the int value specified by "value" is in range specified by argv[]
 * param:
 *   value  -- a int value, string format
 *   argv[] -- a int value range, string format, ex. {"0", "100"}. it has only two element
 * return:
 *   0 -- not
 *   1 -- yes
 */
inline int
valid_range(char *value, char *argv[])
{
	int n, start, end;
	
	n = atoi(value);
	start = atoi(argv[0]);
	end = atoi(argv[1]);
	
	if (n < start || n > end) {
		return 0;
	}
	
	return 1;
}

/*
 * check whether a string length is in a int range
 * param:
 *   value  -- a string
 *   argv[] -- a int value range, string format, ex. {"4", "6"}. it has only two element.
 * return:
 *   0  -- the length of the string is not in the range
 *   1  -- in the range
 */
inline int
valid_name(char *value, char *argv[])
{
	int n, min, max;
	
	n = strlen(value);
	min = atoi(argv[0]);
	max = atoi(argv[1]);
	
	if (n > max || n < min) {
		return 0;
	}
	
	return 1;
}

/*
 * Convert Ethernet address string representation to binary data
 * @param	a	string in xx:xx:xx:xx:xx:xx notation
 * @param	e	binary data
 * @return	TRUE if conversion was successful and FALSE otherwise
 */
int
ether_atoe(const char *a, unsigned char *e)
{
	char *c = (char *) a;
	int i = 0;

	memset(e, 0, ETHER_ADDR_LEN);
	for (;;) {
		e[i++] = (unsigned char) strtoul(c, &c, 16);
		if (!*c++ || i == ETHER_ADDR_LEN)
			break;
	}
	return (i == ETHER_ADDR_LEN);
}

/*
 * Convert Ethernet address binary data to string representation
 * @param	e	binary data
 * @param	a	string in xx:xx:xx:xx:xx:xx notation
 * @return	a
 */
char *
ether_etoa(const unsigned char *e, char *a)
{
	char *c = a;
	int i;

	for (i = 0; i < ETHER_ADDR_LEN; i++) {
		if (i)
			*c++ = ':';
		c += sprintf(c, "%02X", e[i] & 0xff);
	}
	return a;
}

/*
 * chech whether the mac is valid or not
 * return:
 *   0 -- invalid
 *   1 -- valide
 */
int checkmac(const char *mac)
{
	char data[64] = {0}, *token;
	int field = 0, t = 0;

	strcpy(data, mac);
	token = strtok(data, ":");
	while(token) {
		field = strtol(token, NULL, 16);
		if( field < 0 || field > 255 )
			return 0;

		t++;
		token = strtok(NULL, ":");
	}

	return t == ETHER_ADDR_LEN;
}

void dump_uri(request * req, char *uri, char *query, char *name[], char *value[], char *title)
{
	SQUASH_KA(req);
	req->response_status = R_REQUEST_OK;
	req_write(req, "HTTP/1.0 200 OK\r\n");
	print_http_headers(req);
	req_write(req, "Content-Type: " "text/html; charset=ISO-8859-1" "\r\n\r\n"); /* terminate header */

	char buffer[MAX_HEADER_LENGTH] = {0}, *p = buffer;
	p += sprintf(p, "URI = %s\r\nQuery = %s\r\n\r\n", uri, query);

	int i;
	for(i = 0; name[i] && value[i] ; i++ )
		p += sprintf(p, "name  = %s\r\nvalue = %s\r\n\r\n", name[i], value[i]);

	req_write(req,	"<HTML><HEAD><TITLE>");
	req_write(req, title);
	req_write(req, "</TITLE></HEAD>\n<BODY><H1>");
	req_write(req, title);
	req_write(req, "</H1>\n");
	req_write(req, buffer);
	req_write(req,"</BODY></HTML>\n");
	req_flush(req);
}

void send_error_response(request *req, char *fmt, ...)
{
	char data[4*1024] = {0}, cl[64] = {0};
	va_list ap;

	va_start(ap, fmt);
	int t = vsprintf (data, fmt, ap);
	va_end(ap);

	sprintf(cl, "Content-length: %d\r\n\r\n", t);

	SQUASH_KA(req);
        req->response_status = R_REQUEST_OK;
        req_write(req, "HTTP/1.0 200 OK\r\n");
	req_write(req, "Content-Type: text/plain\r\n");
	req_write(req, cl);
	req_write(req, data);
	req_flush(req);
}

void print_header(request *req)
{
        SQUASH_KA(req);
        req->response_status = R_REQUEST_OK;
        req_write(req, "HTTP/1.0 200 OK\r\n");
	req_write(req, "Content-Type: text/plain\r\n");
}

int echo(char *fmt, ...)
{
	static FILE *fp = NULL;
	if( !fp ) {
		fp = fopen("/dev/console", "w");
		if( !fp )
			return 0;
	}

	va_list ap;
	va_start(ap, fmt);
	int t = vfprintf (fp, fmt, ap);
	va_end(ap);

	return t;
}

/*
 * send_error_response result code:
 *  0  -- all done
 *  1  -- notice information
 *  2  -- too few or too many parameters
 *  3  -- invalide param
 *
 */
void viasat_admin(request *req)
{
	char query[MAX_HEADER_LENGTH] = {0}, *type = 0, *mac = 0, pmac[32] = {0}; // suport max 8 name value pairs
	unsigned char bmac[6] = {0};

	strcpy(query, req->query_string);
	type = query;

	if( !(mac = strchr(query, '&')) ) {
		send_error_response(req, "Result: 2\r\nToo few Parameters.");
		echo("Too few Parameters.");
		return ;
	}

	*mac++ = 0;
	echo("type: %s, mac: %s\n", type, mac);

	if( strchr(mac, '&') ) {
		send_error_response(req, "Result: 2\r\nToo Many Parameters.");
		echo("Too Many Parameters.");
		return ;
	}

	/*
	 *    name         value
	 * 0  ap           1 -- AP, 0 -- Client
	 * 1  mac          hex format: xx:xx:xx:xx:xx:xx
	 *
	 */
	if( !valid_choice(type, ARGV("ap", "client", NULL)) ) {
		send_error_response(req, "Result: 3\r\nInvalid Parameters: type=%s", type);
		echo("Invalid Type: %s\n", type);
		return ;
	}

	if( !checkmac(mac) ) {
		send_error_response(req, "Result: 3\r\nInvalid Parameters: mac=%s", mac);
		echo("Invalid MAC: %s\n", mac);
		return ;
	}

	/*
	 * AP MAC:     xx:xx:xx:xx:xx:AA
	 * Client MAC: xx:xx:xx:xx:xx:BB
	 *
	 * AA Must lower than BB
	 */
	ether_atoe(mac, bmac);

	echo("Parameters is OK, now we setup...");
	if( strcmp(type, "ap") == 0 ) { // AP
		bmac[5] += 1;
		nvram_set("et0ipaddr", "192.168.0.27");
	} else  { // Client
		bmac[5] -= 1;
		nvram_set("et0ipaddr", "192.168.0.28");
	}

	send_error_response(req, "Result: 1\r\n%s will Setup:\r\nMAC=%s\r\nPeer MAC=%s\r\nIP=%s", 
		nvram_safe_get("netbiosname"), mac, ether_etoa(bmac, pmac), nvram_safe_get("et0ipaddr"));
	nvram_set("et0macaddr", mac);
	nvram_set("il0macaddr", mac);
	nvram_set("et0peermac", pmac);

	nvram_set("restore_defaults", "1");
	nvram_commit();
	send_error_response(req, "Result: 0\r\nOperation is OK. Restoring and Rebooting...");
	echo("Done.\n");

	sleep(1);
	system("/sbin/reboot");

	return ;
}
