/*
 * Miscellaneous services
 *
 * Copyright 2006, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#include <bcmnvram.h>
#include <netconf.h>
#include <shutils.h>
#include <rc.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>


#define assert(a)

#ifdef __CONFIG_NAT__
static char 
*make_var(char *prefix,int index,char *name)
{
	static char buf[100];
	
	assert(prefix);
	assert(name);
	
	if (index)
		snprintf(buf,sizeof(buf),"%s%d%s",prefix,index,name);
	else
		snprintf(buf,sizeof(buf),"%s%s",prefix,name);
	return buf;
}
int
start_dhcpd(void)
{
	FILE *fp;
	char name[100];
	char word[32], *next;
	char dhcp_conf_file[128];
	char dhcp_lease_file[128];
	char dhcp_ifnames[128];
	int index;

	if (nvram_match("router_disable", "1") )
		return 0;

	/* Setup individual dhcp severs for the bridge and the every unbridged interface */
	   
	snprintf(dhcp_ifnames,sizeof(dhcp_ifnames),"%s",nvram_safe_get("lan_ifname"));
	
	if (nvram_get("unbridged_ifnames"))
		foreach(word,nvram_safe_get("unbridged_ifnames"),next){

			snprintf(dhcp_ifnames,sizeof(dhcp_ifnames),
				"%s %s",dhcp_ifnames,word);
		}
	
	index=0;
	foreach(word, dhcp_ifnames, next){
	
		if (strstr(word,"br0") ) 
			index = 0 ;
		else
			index = get_ipconfig_index(word);
		
		/* Skip interface if no valid config block is found */
		if (index < 0) continue;
		
		if (nvram_invmatch(make_var("lan",index,"_proto"),"dhcp"))
			continue;
			
		dprintf("%s %s %s %s\n",
			nvram_safe_get(make_var("lan",index,"_ifname")),
			nvram_safe_get(make_var("dhcp",index,"_start")),
			nvram_safe_get(make_var("dhcp",index,"_end")),
			nvram_safe_get(make_var("lan",index,"_lease")));

		/* Touch leases file */
		sprintf(dhcp_lease_file,"/tmp/udhcpd%d.leases",index);
		if (!(fp = fopen(dhcp_lease_file, "a"))) {
			perror(dhcp_lease_file);
			return errno;
		}
		fclose(fp);

		/* Write configuration file based on current information */
		sprintf(dhcp_conf_file,"/tmp/udhcpd%d.conf",index);
		if (!(fp = fopen(dhcp_conf_file, "w"))) {
			perror(dhcp_conf_file);
			return errno;
		}
		fprintf(fp, "pidfile /var/run/udhcpd%d.pid\n",index);
		fprintf(fp, "start %s\n", nvram_safe_get(make_var("dhcp",index,"_start")));
		fprintf(fp, "end %s\n", nvram_safe_get(make_var("dhcp",index,"_end")));
		fprintf(fp, "interface %s\n", word);
		fprintf(fp, "remaining yes\n");
		fprintf(fp, "lease_file /tmp/udhcpd%d.leases\n",index);
		fprintf(fp, "option subnet %s\n", nvram_safe_get(make_var("lan",index,"_netmask")));
		fprintf(fp, "option router %s\n", nvram_safe_get(make_var("lan",index,"_ipaddr")));
		fprintf(fp, "option dns %s\n", nvram_safe_get(make_var("lan",index,"_ipaddr")));
		fprintf(fp, "option lease %s\n", nvram_safe_get(make_var("lan",index,"_lease")));
		snprintf(name, sizeof(name), "%s_wins", nvram_safe_get(make_var("dhcp",index,"_wins")));
		if (nvram_invmatch(name, ""))
			fprintf(fp, "option wins %s\n", nvram_get(name));
		snprintf(name, sizeof(name), "%s_domain", nvram_safe_get(make_var("dhcp",index,"_domain")));
		if (nvram_invmatch(name, ""))
			fprintf(fp, "option domain %s\n", nvram_get(name));
		fclose(fp);

		eval("udhcpd", dhcp_conf_file);
		index++;
	}
	dprintf("done\n");
	return 0;
}

int
stop_dhcpd(void)
{
	char sigusr1[] = "-XX";
	int ret;

/*
* Process udhcpd handles two signals - SIGTERM and SIGUSR1
*
*  - SIGUSR1 saves all leases in /tmp/udhcpd.leases
*  - SIGTERM causes the process to be killed
*
* The SIGUSR1+SIGTERM behavior is what we like so that all current client
* leases will be honorred when the dhcpd restarts and all clients can extend
* their leases and continue their current IP addresses. Otherwise clients
* would get NAK'd when they try to extend/rebind their leases and they 
* would have to release current IP and to request a new one which causes 
* a no-IP gap in between.
*/
	sprintf(sigusr1, "-%d", SIGUSR1);
	eval("killall", sigusr1, "udhcpd");
	ret = eval("killall", "udhcpd");

	dprintf("done\n");
	return ret;
}

int
start_dns(void)
{
	int ret;
	FILE *fp;
	char word[32], *next;
	char dns_ifnames[64],cmdline[128];
	
	if (nvram_match("router_disable", "1"))
		return 0;

	/* Create resolv.conf with empty nameserver list */
	if (!(fp = fopen("/tmp/resolv.conf", "w"))) {
		perror("/tmp/resolv.conf");
		return errno;
	}
	fclose(fp);
	
	/* Setup interface list for the dns relay */
	 
	snprintf(dns_ifnames,sizeof(dns_ifnames),"-i %s",nvram_safe_get("lan_ifname"));
		
	foreach(word, nvram_safe_get("unbridged_ifnames"), next){ 
		if ( get_ipconfig_index(word) < 0 ) continue;
		
		/* Check that the interface is truly up otherwise dnsmasq will 
		 * fail to start. Sometimes this can happen if the virtual interfaces
		 * come up but are not fully functional 
		 */
		 
		if (ifconfig(word, IFUP, NULL, NULL)) continue; 
		snprintf(dns_ifnames,sizeof(dns_ifnames),"%s -i %s",dns_ifnames, word);		
	}
	
	/* Start the dns relay */
	snprintf(cmdline,sizeof(cmdline),"/usr/sbin/dnsmasq -h -n %s -r /tmp/resolv.conf",dns_ifnames);
	ret = system(cmdline);

	
	dprintf("done\n");
	return ret;
}	

int
stop_dns(void)
{
	int ret = eval("killall", "dnsmasq");
	
	/* Remove resolv.conf */
	unlink("/tmp/resolv.conf");

	dprintf("done\n");
	return ret;
}	
#endif	/* __CONFIG_NAT__ */

int
start_httpd(void)
{
	int ret;

	chdir("/www");
	ret = eval("httpd");
	chdir("/");

	dprintf("done\n");
	return ret;
}

int
stop_httpd(void)
{
	int ret = eval("killall", "httpd");

	dprintf("done\n");
	return ret;
}

#ifdef __CONFIG_NAT__
int
start_upnp(void)
{
	char *wan_ifname;
	int ret;
	char var[100], prefix[] = "wanXXXXXXXXXX_";

	if (!nvram_invmatch("upnp_enable", "0"))
		return 0;
	
	ret = eval("killall", "-SIGUSR1", "upnp");
	if (ret != 0) {
	    snprintf(prefix, sizeof(prefix), "wan%d_", wan_primary_ifunit());
	    wan_ifname = nvram_match(strcat_r(prefix, "proto", var), "pppoe") ? 
					nvram_safe_get(strcat_r(prefix, "pppoe_ifname", var)) :
					nvram_safe_get(strcat_r(prefix, "ifname", var));
	    ret = eval("upnp", "-D",
		       "-L", nvram_safe_get("lan_ifname"),
		       "-W", wan_ifname);

	}
	dprintf("done\n");
	return ret;
}

int
stop_upnp(void)
{
	int ret = 0;

	if (nvram_match("upnp_enable", "0"))
	    ret = eval("killall", "upnp");

	dprintf("done\n");
	return ret;
}
#endif	/* __CONFIG_NAT__ */



int
start_nas(char *type)
{
	char cfgfile[64];
	char pidfile[64];
	if (!type || !*type)
		type = "lan";
	/* Configure NAS on the bridged interface */
	snprintf(cfgfile, sizeof(cfgfile), "/tmp/nas.%s.conf", type);
	snprintf(pidfile, sizeof(pidfile), "/tmp/nas.%s.pid", type);
	{
		char *argv[] = {"nas", cfgfile, pidfile, type, NULL};
		pid_t pid;
		system("/usr/sbin/nas /tmp/nas.lan.conf /tmp/nas.lan.pid lan &");
		//_eval(argv, NULL, 0, &pid);
		dprintf("done\n");
	}
	
	return 0;
}

/* Start NAS for the guest interfaces */
int
start_guest_nas(void)
{
	char *unbridged_interfaces;
	char *next;
	char name[IFNAMSIZ],lan[IFNAMSIZ];
	int index;
		
	unbridged_interfaces = nvram_get("unbridged_ifnames");
	
	if (unbridged_interfaces)
		foreach(name,unbridged_interfaces,next){
			index = get_ipconfig_index(name);
			if (index < 0) 
				continue;
			snprintf(lan,sizeof(lan),"lan%d",index);
			start_nas(lan);
		}

	return 0;
}
int
stop_nas(void)
{
	FILE *fp;
	int ret = 0;
	char buf[10];

	if(!(fp=fopen("/tmp/nas.lan.pid", "r"))){
		printf("No nas process running\n");
		return -1;
	}
	fgets(buf, sizeof(buf), fp);
	//printf("The nas pid is :%d\n", atoi(buf));
	ret = eval("kill", "-9", buf);

	eval("killall", "nas");

	dprintf("done\n");
	return ret;
}

int
start_ntpc(void)
{
	char *servers = nvram_safe_get("ntp_server");
	
	if (strlen(servers)) {
		char *nas_argv[] = {"ntpclient", "-h", servers, "-i", "3", "-l", "-s", NULL};
		pid_t pid;

		_eval(nas_argv, NULL, 0, &pid);
	}
	
	dprintf("done\n");
	return 0;
}

int
stop_ntpc(void)
{
	int ret = eval("killall", "ntpclient");

	dprintf("done\n");
	return ret;
}

char * get_netbiosname(void)
{
	char namebuf[14], *macaddr, *netbiosname;
	int s,i = 0;
	
	//struct ifreq ifr;
	//int f_failed = 0;
	//u_char macaddr[6];
	
	netbiosname = nvram_safe_get("netbiosname");
	if ( strlen(netbiosname) == 0 ) {

		macaddr = nvram_safe_get("et0macaddr");
		if ( strlen(macaddr) < 17 )
			macaddr = "00:00:00:00:00:00";
		
#if 0
		/* Get the mac address of eth0 */
		strncpy(ifr.ifr_name, "eth0", sizeof(ifr.ifr_name));
		s = socket(AF_PACKET, SOCK_PACKET, 0);
		if ( ioctl(s, SIOCGIFHWADDR, &ifr) < 0 ) {
			fprintf(stderr, "%s:%d SIOCGIFHWADDR : %s", __FILE__, __LINE__, strerror(errno));
			f_failed = 1;
		}
		if ( ifr.ifr_hwaddr.sa_family != ARPHRD_ETHER ) {
			fprintf(stderr, "%s:%d Interface is not ethernet", __FILE__, __LINE__);
			f_failed = 1;
		}
		if (f_failed == 0)
		{
			bcopy(ifr.ifr_hwaddr.sa_data, macaddr, sizeof(macaddr));
		}
#endif
		snprintf(namebuf, sizeof(namebuf), "%s%c%c%c%c%c%c", "netgear", macaddr[9], macaddr[10], macaddr[12], macaddr[13], macaddr[15], macaddr[16]);
		nvram_set("netbiosname", namebuf);
		netbiosname = nvram_safe_get("netbiosname");
	}

	return netbiosname;
}

int start_wins(void)
{
	const char netbios_cfg[] = "/var/etc/netbios.conf";
	FILE *fp;
	char *netbiosname;
	int ret;

	/* Write netbios configuration file based on current information */
	if (!(fp = fopen(netbios_cfg,"w"))) {
		perror(netbios_cfg);
		return errno;
	}
	netbiosname = get_netbiosname();
	fprintf(fp, "%s\n", get_netbiosname() );
	dprintf( "netbiosname : %s\n", netbiosname );
	fclose(fp);

	/* run wins */
	ret = eval("wins");
	dprintf("done\n");
	return ret;
}

int stop_wins(void)
{
	int ret = eval("killall", "wins");

	dprintf("done\n");
	return ret;
}

int copy_command(void)
{
	int ret;

	ret = eval("cp", "/sbin/lockfile","/tmp");
	ret = eval("chmod","755", "/tmp/lockfile");
	dprintf("done\n");
	return ret;
}

int start_services(void)
{
  //start_httpd();
#ifdef __CONFIG_NAT__
	start_dns();
	start_dhcpd();
	start_upnp();
#endif	/* __CONFIG_NAT__ */
	start_nas("lan");
	start_guest_nas(); 
	start_wins();
	dprintf("done\n");
	return 0;
}

int stop_services(void)
{
	stop_nas();
#ifdef __CONFIG_NAT__
	stop_upnp();
	stop_dhcpd();
	stop_dns();
#endif	/* __CONFIG_NAT__ */
	//stop_httpd();
	stop_wins();
#ifdef __CONFIG_MONITORD__
	eval("killall", "mond");
#endif

	dprintf("done\n");
	return 0;
}
